/*
 * Sample.c
 *
 * (C) Copyright Argonaut Software. 1994-1996.  All rights reserved.
 *
 * A sample Windows application
 */
#include <windows.h>
#include <brender.h>

/*
 * Global variables
 */
int angle = 0;
int window_x = 0;
int window_y = 0;
int window_width= 400;
int window_height = 300;
int background_colour;

br_boolean fullscreen = BR_FALSE;
br_pixelmap *screen, *offscreen;
br_pixelmap *depth, *palette;
br_actor *world, *light, *camera, *object;
br_material *material;

HWND hwnd;
WNDCLASS wclass;

/*
 * Diagnostic handler functions
 */
static void BR_CALLBACK BrWinWarning(char *message)
{
    MessageBox(NULL, message, "WARNING", MB_OK | MB_TASKMODAL);
	ExitProcess(10);
}

static void BR_CALLBACK BrWinFailure(char *message)
{
    MessageBox(NULL, message, "FAILURE", MB_OK | MB_TASKMODAL);
	ExitProcess(10);
}

/*
 * Diagnostic handler
 */
static br_diaghandler BrWinDiagHandler = {
	"Windows Diagnostic Handler",
	BrWinWarning,
	BrWinFailure,
};

/*
 * Get pixelmaps
 */
static void GetPixelmaps(void)
{
	br_error result;
	br_camera *camera_data;

	/*
	 * Request screen pixelmap from device
	 */
	if(!fullscreen) {
		result = BrDevBeginVar(&screen, "DDRAW",
							   BRT_WINDOW_HANDLE_H, hwnd,
							   BRT_WINDOW_BUFFER_METHOD_T, BRT_DIRECTDRAW,
							   BRT_WINDOW_FULLSCREEN_B, BR_FALSE,
							   BRT_PIXEL_BITS_I32, 15,
							   BR_NULL_TOKEN);
	} else {
		result = BrDevBeginVar(&screen, "DDRAW",
							   BRT_WINDOW_HANDLE_H, hwnd,
							   BRT_WINDOW_BUFFER_METHOD_T, BRT_DIRECTDRAW,
							   BRT_WINDOW_FULLSCREEN_B, BR_TRUE,
							   BRT_WIDTH_I32, 640,
							   BRT_HEIGHT_I32, 480,
							   BRT_PIXEL_BITS_I32, 15,
							   BR_NULL_TOKEN);
	}

	if(result != BRE_OK)
		BR_FAILURE("Could not get screen pixelmap");

	/*
	 * Set palette if indexed pixelmap
	 */
	if(screen->type == BR_PMT_INDEX_8)
		BrPixelmapPaletteSet(screen, palette);

	/*
	 * Match offscreen pixelmap from screen
	 */
	offscreen = BrPixelmapMatch(screen, BR_PMMATCH_OFFSCREEN);
	if(offscreen == NULL)
		BR_FAILURE("Could not match offscreen pixelmap");

	/*
	 * Adjust origin
	 */
	offscreen->origin_x = offscreen->width / 2;
	offscreen->origin_y = offscreen->height / 2;

	/*
	 * Adjust camera aspect ratio
	 */
	camera_data = (br_camera *)camera->type_data;
	camera_data->aspect = BR_DIV(BrIntToScalar(offscreen->width), BrIntToScalar(offscreen->height));

	/*
	 * Match depth pixelmap from offscreen
	 */
	depth = BrPixelmapMatch(offscreen, BR_PMMATCH_DEPTH_16);
	if(depth == NULL)
		BR_FAILURE("Could not match depth pixelmap");

	/*
	 * Initialise renderer
	 */
	BrZbBegin(offscreen->type, depth->type);
}

/*
 * Release pixelmaps
 */
static void ReleasePixelmaps(void)
{
	/*
	 * Free pixelmaps
	 */
	BrPixelmapFree(depth);
	BrPixelmapFree(offscreen);
	BrPixelmapFree(screen);

	/*
	 * Clear pointers
	 */
	depth = NULL;
	offscreen = NULL;
	screen = NULL;
}

/*
 * Adjust window size
 */
static void SizeWindow(void)
{
	int x, y, width, height;

	if(fullscreen) {
		/*
		 * Set to full screen
		 */
		x = 0;
		y = 0;
		width = GetSystemMetrics(SM_CXSCREEN);
		height = GetSystemMetrics(SM_CYSCREEN);
	} else {
		/*
		 * Set to original window size
		 */
		x = window_x;
		y = window_y;
		width = window_width;
		height = window_height;
	}

	SetWindowPos(hwnd, HWND_TOP, x, y, width, height, SWP_NOCOPYBITS);
}

/*
 * Window procedure
 */
static LRESULT CALLBACK WindowProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	int width, height;
	br_camera *camera_data;
	RECT rect;

	switch(message) {

	/*
	 * Post the WM_QUIT message
	 */
	case WM_DESTROY:
		PostQuitMessage(0);
		break;

	/*
	 * Background is not erased
	 */
	case WM_ERASEBKGND:
		return 1;

	/*
	 * Window size has changed
	 */
	case WM_SIZE:
		/*
		 * Ignore message on window creation
		 */
		if(lParam == 0)
			break;

		/*
		 * Ignore change to fullscreen
		 */
		if(fullscreen)
			break;

		/*
		 * Extract client area dimensions
		 */
		width = (int)LOWORD(lParam);
		height = (int)HIWORD(lParam);

		/*
		 * Remember window dimensions
		 */
		GetWindowRect(hwnd, &rect);
		window_x = rect.left;
		window_y = rect.top;
		window_width = rect.right - rect.left;
		window_height = rect.bottom - rect.top;

		/*
		 * Resize screen pixelmap
		 */
		if(screen != NULL)
			BrPixelmapResize(screen, width, height);

		/*
		 * Resize offscreen pixelmap
		 */
		if(offscreen != NULL) {
			BrPixelmapResize(offscreen, width, height);
			offscreen->origin_x = offscreen->width / 2;
			offscreen->origin_y = offscreen->height / 2;
		}

		/*
		 * Resize depth pixelmap
		 */
		if(depth != NULL)
			BrPixelmapResize(depth, width, height);

		/*
		 * Adjust camera aspect ratio
		 */
		camera_data = (br_camera *)camera->type_data;
		camera_data->aspect = BR_DIV(BrIntToScalar(width), BrIntToScalar(height));

		break;

	/*
	 * Window has moved
	 */
	case WM_MOVE:
		if(!fullscreen) {
			/*
			 * Remember new position
			 */
			window_x = (int)LOWORD(lParam);
			window_y = (int)HIWORD(lParam);
		}
		break;

	/*
	 * Check for ESC keypress
	 */
	case WM_KEYDOWN:
		switch(wParam) {
		case VK_ESCAPE:
			/*
			 * Toggle fullscreen flag
			 */
			fullscreen = 1 - fullscreen;

			/*
			 * Release existing pixelmaps
			 */
			ReleasePixelmaps();

			/*
			 * Adjust window size appropriately
			 */
			SizeWindow();

			/*
			 * Get new pixelmaps
			 */
			GetPixelmaps();
		}
		break;

	/*
	 * System palette has changed
	 */
	case WM_PALETTECHANGED:
		if((HWND)wParam == hWnd)
			break;

	/*
	 * Realise palette
	 */
	case WM_QUERYNEWPALETTE:
		if(screen != NULL)
			BrPixelmapPaletteSet(screen, palette);
		break;

	default:
		return DefWindowProc(hWnd, message, wParam, lParam);
	}

	return 0;
}

/*
 * Initialisation
 */
static BOOL AppBegin(HINSTANCE hInstance, int nCmdShow)
{
	/*
	 * Register a new window class
	 */
	wclass.style = CS_HREDRAW | CS_VREDRAW;
	wclass.lpfnWndProc = WindowProc;
	wclass.cbClsExtra = 0;
	wclass.cbWndExtra = 0;
	wclass.hInstance = hInstance;
	wclass.hIcon = LoadIcon(hInstance, IDI_APPLICATION);
	wclass.hCursor = LoadCursor(NULL, IDC_ARROW);
	wclass.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
	wclass.lpszMenuName = NULL;
	wclass.lpszClassName = "SampleClass";
	RegisterClass(&wclass);

	/*
	 * Create a window
	 */
	hwnd = CreateWindowEx(0, "SampleClass", "Sample", WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT, CW_USEDEFAULT, 400, 300, NULL, NULL, hInstance, NULL);

	if(hwnd == NULL)
		BR_FAILURE("Could not create window");

	/*
	 * Initialise BRender
	 */
	BrBegin();

	/*
	 * Set diagnostic handler
	 */
	BrDiagHandlerSet(&BrWinDiagHandler);

	/*
	 * Create a simple world
	 */
	world = BrActorAllocate(BR_ACTOR_NONE, NULL);

	/*
	 * Camera
	 */
	camera = BrActorAdd(world, BrActorAllocate(BR_ACTOR_CAMERA, NULL));
	BrVector3Set(&camera->t.t.translate.t, BR_SCALAR(0.0), BR_SCALAR(0.0), BR_SCALAR(5.0));

	/*
	 * Light
	 */
	light = BrActorAdd(world, BrActorAllocate(BR_ACTOR_LIGHT, NULL));
	BrLightEnable(light);

	/*
	 * Default model
	 */
	object = BrActorAdd(world, BrActorAllocate(BR_ACTOR_MODEL, NULL));
	object->t.type = BR_TRANSFORM_EULER;
	object->t.t.euler.e.order = BR_EULER_ZXY_S;

	/*
	 * Material (with base and range to match palette)
	 */
	material = BrMaterialAllocate(NULL);
	material->colour = BR_COLOUR_RGB(0,255,0);
	material->index_base = 74;
	material->index_range = 31;
	material->flags = BR_MATF_LIGHT;
	object->material = BrMaterialAdd(material);

	/*
	 * Load palette
	 */
	palette = BrPixelmapLoad("winpal.pix");
	if(palette == NULL)
		BR_FAILURE("Could not load palette pixelmap");

	/*
	 * Get screen, offscreen and depth pixelmaps
	 */
	GetPixelmaps();

	/*
	 * Show and update window
	 */
	ShowWindow(hwnd, nCmdShow);
	UpdateWindow(hwnd);

	return TRUE;
}

/*
 * Display pixelmap type
 */
static void DisplayPixelmapType(br_pixelmap *pmap)
{
	br_int_16 x,y;

	x = 2 - pmap->width / 2;
	y = 2 - pmap->height / 2;

	switch(offscreen->type) {
	case BR_PMT_INDEX_8:
		BrPixelmapText(offscreen, x, y, 41, NULL, "Pixelmap type: BR_PMT_INDEX_8");
		break;
	case BR_PMT_RGB_555:
		BrPixelmapText(offscreen, x, y, 0x7fff, NULL, "Pixelmap type: BR_PMT_RGB_555");
		break;
	case BR_PMT_RGB_565:
		BrPixelmapText(offscreen, x, y, 0xffff, NULL, "Pixelmap type: BR_PMT_RGB_565");
		break;
	case BR_PMT_RGB_888:
		BrPixelmapText(offscreen, x, y, 0xffffff, NULL, "Pixelmap type: BR_PMT_RGB_888");
		break;
	}
}

/*
 * Update the world and render it
 */
static void Step(void)
{
	/*
	 * Determine background colour
	 */
	background_colour = 0;
	if(offscreen->type == BR_PMT_INDEX_8)
		background_colour = 10;

	/*
	 * Clear offscreen and depth pixelmaps
	 */
	BrPixelmapFill(offscreen, background_colour);
	BrPixelmapFill(depth, 0xffff);

	/*
	 * Render the scene
	 */
	BrZbSceneRender(world, camera, offscreen, depth);

	/*
	 * Display offscreen pixelmap type
	 */
	DisplayPixelmapType(offscreen);

	/*
	 * Update the world
	 */
	object->t.t.euler.e.a = BR_ANGLE_DEG(angle);
	object->t.t.euler.e.b = BR_ANGLE_DEG(angle * 2);
	object->t.t.euler.e.c = BR_ANGLE_DEG(angle * 3);
	if(++angle == 360) angle = 0;

	/*
	 * Double buffer
	 */
	BrPixelmapDoubleBuffer(screen, offscreen);
}

/*
 * WinMain()
 *
 * Windows entry point
 */
int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	MSG msg;

	/*
	 * Initialisation
	 */
	if(!AppBegin(hInstance, nCmdShow))
		return FALSE;

	/*
	 * Message loop
	 */
	for(;;) {
		if(PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE)) {
			if(!GetMessage(&msg, NULL, 0, 0))
				break;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		} else {
			Step();
		}
	}

	return msg.wParam;
}
